/*
 * Licensed under BSD license.  See LICENCE.TXT  
 *
 * Produced by:	Jeff Lait
 *
 *      	7DRL Development
 *
 * NAME:        gfxengine.cpp ( Live Once Library, C++ )
 *
 * COMMENTS:
 */

#include "gfxengine.h"
#include "ptrlist.h"
#include "buf.h"

#include "gfx_sdl.h"

#define INVOKEGFX(func) gfxsdl_##func

#include <iostream>
#include <fstream>
using namespace std;

#include <assert.h>

// Screen Listing
u8	glbScreenCharMap[SCR_HEIGHT][SCR_WIDTH];
int	glbScreenAttrMap[SCR_HEIGHT][SCR_WIDTH];
TILE_NAMES glbScreenTileMap[MAX_LAYER][SCR_HEIGHT/4][SCR_WIDTH/4];
int glbScreenTileMapOffX[MAX_LAYER][SCR_HEIGHT/4][SCR_WIDTH/4];
int glbScreenTileMapOffY[MAX_LAYER][SCR_HEIGHT/4][SCR_WIDTH/4];
int glbScreenTileLight[SCR_HEIGHT/4][SCR_WIDTH/4];
int glbScreenTileLevel[SCR_HEIGHT/4][SCR_WIDTH/4];
bool	glbScreenDirty = false;

// There is one true endian style.
bool	glbEvilEndian = false;

GFX_FadeMode	glbFadeMode;
u8		glbFadeAmount;
int		glbScrollOffsetX = 0;
int		glbScrollOffsetY = 0;
PTRLIST<gfx_bmpoverlay *>	glbOverlayList;

void
gfx_init()
{
    INVOKEGFX(init)();

    // Determine our endianness.
    volatile unsigned	short	 s;
    u8			*c;

    s = 0;
    c = (u8 *) &s;
    *c = 1;

    for (int y = 0; y < SCR_HEIGHT/4; y++)
    {
	for (int x = 0; x < SCR_WIDTH/4; x++)
	{
	    gfx_settile_light(x, y, 255);
	    gfx_settile_level(x, y, 0);
	}
    }

    // The short will now either be 256 or 1.
    if (s == 1)
	glbEvilEndian = false;
    else
	glbEvilEndian = true;
}

void
gfx_shutdown()
{
    INVOKEGFX(shutdown)();
}

void 
gfx_setfade(GFX_FadeMode mode, u8 amount)
{
    glbFadeMode = mode;
    glbFadeAmount = amount;
    glbScreenDirty = true;
}

void
gfx_setscrolloffset(int dx, int dy)
{
    glbScrollOffsetX = dx;
    glbScrollOffsetY = dy;
    glbScreenDirty = true;
}

void
gfx_getscrolloffset(int &dx, int &dy)
{
    dx = glbScrollOffsetX;
    dy = glbScrollOffsetY;
}

void
gfx_shake(int duration)
{
    int		dx = 0;
    int		dy = 0;
    while (duration --> 0)
    {
	dx += rand_choice(11)-5;
	dy += rand_choice(11)-5;
	gfx_setscrolloffset(dx, dy);
	gfx_update();
	gfx_delay(1);
    }

    while (dx || dy)
    {
	dx -= SIGN(dx);
	dy -= SIGN(dy);
	gfx_setscrolloffset(dx, dy);
	gfx_update();
	gfx_delay(1);
    }
}

void
gfx_fadeto(GFX_FadeMode mode, bool increase)
{
    if (increase)
    {
	for (int amt = 0; amt < 255; amt += 16)
	{
	    gfx_setfade(mode, amt);
	    gfx_update();
	    gfx_delay(2);
	}
	gfx_setfade(mode, 255);
	gfx_update();
	gfx_delay(2);
    }
    else
    {
	for (int amt = 255; amt > 0; amt -= 16)
	{
	    gfx_setfade(mode, amt);
	    gfx_update();
	    gfx_delay(2);
	}
	gfx_setfade(mode, 0);
	gfx_update();
	gfx_delay(2);
    }
}

void
gfx_delay(int frames)
{
    INVOKEGFX(delay)(frames);
}

void
gfx_printchar(int x, int y, u8 c, ATTR_NAMES attr)
{
    assert(x >= 0 && x < SCR_WIDTH);
    assert(y >= 0 && y < SCR_HEIGHT);
    glbScreenCharMap[y][x] = c;
    glbScreenAttrMap[y][x] = attr;
    glbScreenDirty = true;
}

void
gfx_settile(int layer, int x, int y, TILE_NAMES tile, int offx, int offy)
{
    assert(y >= 0 && y < SCR_HEIGHT/4);
    assert(x >= 0 && x < SCR_WIDTH/4);
    assert(layer >= 0 && layer < MAX_LAYER);
    glbScreenTileMap[layer][y][x] = tile;
    glbScreenTileMapOffX[layer][y][x] = offx;
    glbScreenTileMapOffY[layer][y][x] = offy;
    glbScreenDirty = true;
}

TILE_NAMES
gfx_gettile(int layer, int x, int y, int &offx, int &offy)
{
    assert(y >= 0 && y < SCR_HEIGHT/4);
    assert(x >= 0 && x < SCR_WIDTH/4);
    assert(layer >= 0 && layer < MAX_LAYER);
    offx = glbScreenTileMapOffX[layer][y][x];
    offy = glbScreenTileMapOffY[layer][y][x];
    return glbScreenTileMap[layer][y][x];
}

TILE_NAMES
gfx_gettile(int layer, int x, int y)
{
    assert(y >= 0 && y < SCR_HEIGHT/4);
    assert(x >= 0 && x < SCR_WIDTH/4);
    assert(layer >= 0 && layer < MAX_LAYER);
    return glbScreenTileMap[layer][y][x];
}

void
gfx_settile_light(int x, int y, int light)
{
    assert(y >= 0 && y < SCR_HEIGHT/4);
    assert(x >= 0 && x < SCR_WIDTH/4);

    glbScreenTileLight[y][x] = light;
}

int
gfx_gettile_light(int x, int y)
{
    assert(y >= 0 && y < SCR_HEIGHT/4);
    assert(x >= 0 && x < SCR_WIDTH/4);

    return glbScreenTileLight[y][x];
}


void
gfx_settile_level(int x, int y, int level)
{
    assert(y >= 0 && y < SCR_HEIGHT/4);
    assert(x >= 0 && x < SCR_WIDTH/4);
    assert(level >= 0 && level < MAX_LEVELS);

    glbScreenTileLevel[y][x] = level;
}

int
gfx_gettile_level(int x, int y)
{
    assert(y >= 0 && y < SCR_HEIGHT/4);
    assert(x >= 0 && x < SCR_WIDTH/4);

    return glbScreenTileLevel[y][x];
}

void
gfx_printstring(int x, int y, const char *s, ATTR_NAMES attr)
{
    if (y <0 || y >= SCR_HEIGHT)
	return;

    while (*s && x < 0)
    {
	s++;
	x++;
    }
    while (*s && x < SCR_WIDTH)
    {
	gfx_printchar(x, y, *s, attr);
	s++;
	x++;
    }
}

void
gfx_justifystring(int x, int y, JUST_Horizontal jh, JUST_Vertical jv,
		    const char *s, ATTR_NAMES attr)
{
    // Break our string into lines.
    if (!s)
	return;
    char			*copys = strdup(s);
    PTRLIST<char *>		lines;

    lines.append(copys);
    for (char *i = copys; *i; i++)
    {
	if (*i == '\n')
	{
	    *i = '\0';
	    i++;
	    if (*i)
	    {
		lines.append(i);
	    }
	}
    }

    // Now print.
    int			sy = y;
    if (jv == JUST_BOTTOM)
	sy -= lines.entries() -1;

    for (int i = 0; i < lines.entries(); i++)
    {
	int		sx = x;
	int		len = strlen(lines(i));
	if (jh == JUST_RIGHT)
	{
	    sx -= len - 1;
	}
	if (jh == JUST_CENTER)
	{
	    sx -= len/2;
	}
	gfx_printstring(sx, sy, lines(i), attr);
	sy++;
    }

    free(copys);
}

void
gfx_clearbox(int sx, int sy, int w, int h, u8 c, ATTR_NAMES attr)
{
    int		x, y;
    
    for (y = sy; y < sy + h; y++)
    {
	for (x = sx; x < sx + w; x++)
	{
	    gfx_printchar(x, y, c, attr);
	}
    }
}

int
gfx_overlay_create(const u8 *rgb, int w, int h,
		    int x, int y, int scale,
		    int layer,
		    bool transparent)
{
    // Find overlay handle
    int		ovr = 0;
    for ( ; ovr < glbOverlayList.entries(); ovr++)
    {
	if (!glbOverlayList(ovr))
	    break;
    }
    if (ovr >= glbOverlayList.entries())
	ovr = glbOverlayList.append(0);

    gfx_bmpoverlay	*o = new gfx_bmpoverlay;

    o->rgbdata = new u8[w * h * 3];
    memcpy(o->rgbdata, rgb, w * h * 3);
    o->x = x;
    o->y = y;
    o->w = w;
    o->h = h;
    o->layer = layer;
    o->scale = scale;
    o->transparent = transparent;

    glbOverlayList.set(ovr, o);

    glbScreenDirty = true;

    return ovr;
}

void
gfx_overlay_center(int handle)
{
    assert(handle >= 0);
    assert(handle < glbOverlayList.entries());
    assert(glbOverlayList(handle));

    gfx_bmpoverlay	*o = glbOverlayList(handle);
    if (!o)
	return;

    // Don't have access to font height here.
    o->x = (SCR_WIDTH * 8 - o->w * o->scale) / 2;
    o->y = (SCR_HEIGHT * 8 - o->h * o->scale) / 2;
}

void
gfx_overlay_destroy(int handle)
{
    assert(handle >= 0);
    assert(handle < glbOverlayList.entries());
    assert(glbOverlayList(handle));

    gfx_bmpoverlay	*o = glbOverlayList(handle);
    if (!o)
	return;

    delete [] o->rgbdata;
    delete o;
    glbOverlayList.set(handle, 0);
    
    glbScreenDirty = true;
}

void
gfx_getchar(int x, int y, u8 &c, ATTR_NAMES &attr)
{
    assert(x >= 0 && x < SCR_WIDTH);
    assert(y >= 0 && y < SCR_HEIGHT);

    c = glbScreenCharMap[y][x];
    attr = (ATTR_NAMES) glbScreenAttrMap[y][x];
}

void
gfx_update()
{
    if (glbScreenDirty)
    {
	INVOKEGFX(rebuildscreen)();
	glbScreenDirty = false;
    }
}

bool
gfx_isKeyWaiting()
{
    return INVOKEGFX(isKeyWaiting)();
}

void
gfx_clearKeyBuffer()
{
    INVOKEGFX(clearKeyBuffer)();
}

void
gfx_breakKeyRepeat()
{
    INVOKEGFX(breakKeyRepeat)();
}

int
gfx_getKey(bool block)
{
    if (block)
    {
	while (!gfx_isKeyWaiting())
	{
	    INVOKEGFX(awaitEvent)();
	    // Force dirty as we either have a key event or a 
	    // frame event!
	    glbScreenDirty = true;
	    gfx_update();
	}
    }
    return INVOKEGFX(getKey)();
}

//
// Bitmap loading code.
// (Pre 7DRL)
//

static short
gfx_readportableshort(istream &is)
{
    u8		c[2];
    short	result;

    is.read((char *) c, 2);

    if (glbEvilEndian)
    {
	result = c[0] << 8;
	result |= c[1];
    }
    else
    {
	result = c[0];
	result |= c[1] << 8;
    }
    return result;
}

static int
gfx_readportableint(istream &is)
{
    unsigned short	s1, s2;
    int			result;

    s1 = (unsigned short) gfx_readportableshort(is);
    s2 = (unsigned short) gfx_readportableshort(is);

    if (glbEvilEndian)
    {
	result = s1 << 16;
	result |= s2;
    }
    else
    {
	result = s1;
	result |= s2 << 16;
    }

    return result;
}

struct BMPHEAD
{
    char    id[2];
    int	    size;
    int reserved;
    int	    headersize;
    int	    infosize;
    int	    width;
    int	    depth;
    short   biplane;
    short   bits;
    int	    comp;
    int	    imagesize;
    int	    xpelperm;
    int	    ypelperm;
    int	    clrused;
    int	    clrimport;
};

struct RGBQUAD
{
    u8 blue;
    u8 green;
    u8 red;
    u8 filler;
};

u8 *
gfx_loadbitmap(const char *fname, int &w, int &h)
{
    ifstream	    is;
    BMPHEAD	    head;
    int		    i, x, y;

    is.open(fname, ios_base::in | ios_base::binary);

    if (!is)
    {
	cerr << "Failure to open: " << fname << endl;
	return 0;
    }

    is.read(head.id, 2);
    
    head.size = gfx_readportableint(is);
    head.reserved = gfx_readportableint(is);
    head.headersize = gfx_readportableint(is);
    head.infosize = gfx_readportableint(is);
    head.width = gfx_readportableint(is);
    head.depth = gfx_readportableint(is);
    head.biplane = gfx_readportableshort(is);
    head.bits = gfx_readportableshort(is);
    head.comp = gfx_readportableint(is);
    head.imagesize = gfx_readportableint(is);
    head.xpelperm = gfx_readportableint(is);
    head.ypelperm = gfx_readportableint(is);
    head.clrused = gfx_readportableint(is);
    head.clrimport = gfx_readportableint(is);

    if (head.id[0] != 'B' && head.id[1] != 'M')
    {
	cerr << fname << " does not look like a bit map file!" << endl;
	return 0;
    }

    if (head.bits != 32 && head.bits != 24 && head.bits != 8)
    {
	cerr << "Error reading " << fname << endl;
	cerr << "Head.bits is " << head.bits << endl;
	cerr << "Cannot support non 32, 24 or 8 bit colour bitmaps." << endl;
	return 0;
    }

    u8	*result;

    h = head.depth;
    w = head.width;

    result = new u8 [h * w * 3];

    RGBQUAD	palette[256];

    if (head.bits == 8)
    {
	is.read((char *) &palette, sizeof(RGBQUAD) * 256);
    }

    // BMPs are stored bottom up, so we reverse it here.
    for (y = head.depth - 1; y >= 0; y--)
    {
	for (x = 0; x < head.width; x++)
	{
	    u8		red, blue, green, alpha, idx;

	    alpha = 255;
	    if (head.bits == 32)
	    {
		// BMPs are stored BGR.
		is.read((char *) &red, 1);
		is.read((char *) &green, 1);
		is.read((char *) &blue, 1);
		is.read((char *) &alpha, 1);
	    }
	    else if (head.bits == 24)
	    {
		// BMPs are stored BGR.
		is.read((char *) &red, 1);
		is.read((char *) &green, 1);
		is.read((char *) &blue, 1);
	    }
	    else
	    {
		is.read((char *) &idx, 1);
		red = palette[idx].blue;
		green = palette[idx].green;
		blue = palette[idx].red;
	    }
	
	    // Account for endian issues & repack
	    i = x + y * head.width;

	    result[i*3] = red;
	    result[i*3+1] = green;
	    result[i*3+2] = blue;
	}
    }

    return result;
}

u8 *
gfx_loadbitmap_override(const char *basename, int level, int &w, int &h)
{
    ifstream	    is;

    BUF			fname;

    fname.sprintf("../gfx/level%d/%s", level, basename);
    is.open(fname.buffer(), ios_base::in | ios_base::binary);
    if (!is)
    {
	// Try base name.
	fname.sprintf("../gfx/%s", basename);
	is.open(fname.buffer(), ios_base::in | ios_base::binary);
	if (!is)
	{
	    cerr << "Failure to open basename: " << basename << endl;
	    return 0;
	}
	is.close();
	return gfx_loadbitmap(fname.buffer(), w, h);
    }

    return gfx_loadbitmap(fname.buffer(), w, h);
}

void
gfx_getString(int x, int y, ATTR_NAMES attr, char *buf, int maxlen)
{
    int		len;
    int		key;

    len = 0;
    buf[len] = 0;

    if (maxlen + x >= SCR_WIDTH)
	maxlen = SCR_WIDTH - x;

    while (1)
    {
	gfx_update();
	key = gfx_getKey();
	// Ignore non ascii.
	if (key > 255)
	    continue;
	// This should not happen because we block.
	if (!key)
	    continue;

	if (key == '\n' || key == '\r')
	{
	    // Return key! Done!
	    return;
	}

	// If it is backspace, write a blank & reverse.
	if (key == '\b')
	{
	    // Ignore if already empty
	    if (!len)
		continue;
	    // decremen & write a space.
	    len--;
	    buf[len] = '\0';
	    gfx_printchar(x+len, y, ' ', attr);
	    continue;
	}

	// Add the key to the buffer.
	if (len < maxlen-1)
	{
	    buf[len] = key;
	    gfx_printchar(x+len, y, key, attr);
	    len++;
	    buf[len] = '\0';
	}
    }
}
