#include "glbdef.h"
#include "gfxengine.h"
#include <stdio.h>
#include <iostream>
#include <time.h>
using namespace std;

#include "map.h"
#include "atlas.h"
#include "build.h"
#include "mob.h"
#include "panel.h"
#include "text.h"
#include "runes.h"
#include "config.h"

MOB *glbAvatar = 0;
PANEL *glbMsg = 0;
PANEL *glbTalkBox = 0;
PANEL *glbBook = 0;
bool glbVictory = false;

void
redrawWorld()
{
    gfx_clearbox(0, 0, SCR_WIDTH, SCR_HEIGHT, ' ', ATTR_NORMAL);
    if (glbAvatar)
	MAP::render(glbAvatar->pos(), DIR::north(), glbAvatar);
    glbMsg->redraw();
    glbTalkBox->redraw();
    glbBook->redraw();
    gfx_update();
}

void
fadeToBlack()
{
    redrawWorld();
    gfx_fadeto(GFX_FADE_BLACK, false);
    gfx_clearKeyBuffer();
}

void
fadeToWhite()
{
    redrawWorld();
    gfx_fadeto(GFX_FADE_WHITE, false);
    gfx_setfade(GFX_FADE_NONE, 0);
    gfx_clearKeyBuffer();
}

void
fadeFromBlack()
{
    redrawWorld();
    gfx_fadeto(GFX_FADE_BLACK, true);
    gfx_setfade(GFX_FADE_NONE, 0);
    gfx_clearKeyBuffer();
}

void
readBook(const char *text)
{
    glbBook->clear();
    glbBook->appendText(text);

#if 0
    int		height;
    height = (glbBook->getCurLine() + 3) / 4;
    for (int x = 0; x < 10; x++)
    {
	gfx_settile(GFX_LAYER_BOOK, x + 10, 2, TILE_PAPER_TOP);
	for (int y = 3; y < 3 + height; y++)
	{
	    gfx_settile(GFX_LAYER_BOOK, x + 10, y, TILE_PAPER);
	}
	gfx_settile(GFX_LAYER_BOOK, x + 10, 3 + height, TILE_PAPER_BOTTOM);
    }
#endif
    redrawWorld();
    gfx_getKey(true);

#if 0
    for (int x = 0; x < 10; x++)
    {
	for (int y = 2; y < 4 + height; y++)
	{
	    gfx_settile(GFX_LAYER_BOOK, x + 10, y, TILE_NONE);
	}
    }
#endif
    glbBook->clear();
}
void
readBook(BUF buf) { readBook(buf.buffer()); }

void
gba_main()
{
    glbConfig = new CONFIG();
    glbConfig->load("../daystar.cfg");

    gfx_init();
    text_init();

    glbMsg = new PANEL(40, 20);
    glbMsg->move(3, 3);
    glbMsg->setTransparent(true);
    glbTalkBox = new PANEL(100, 1);
    glbTalkBox->move(3, 50);
    glbTalkBox->setTransparent(true);
    glbBook = new PANEL(39, 50);
    glbBook->move(41, 12);
    glbBook->setTransparent(true);
    glbBook->setIndent(1);

    gfx_setfade(GFX_FADE_BLACK, 0);

    glbMsg->move(20, 20);

    // Welcome
    glbMsg->appendText(text_lookup("game", "welcome"));
    redrawWorld();
    gfx_getKey(true);
    glbMsg->erase();
    glbMsg->clear();

    rand_setseed((long) time(0));

    RUNES::getRunes()->init();

    // Register our variables.
    {
	SPELL_NAMES spell;
	FOREACH_SPELL(spell)
	{
	    text_registervariable(glb_spelldefs[spell].words, RUNES::getRunes()->spellRune(spell));
	}
    }

    int		border_overlay, blaze_overlay;
    {
	int		 borderw, borderh;
        u8		*border_rgb = gfx_loadbitmap("../gfx/border.bmp", borderw, borderh);

	if (border_rgb)
	{
	    border_overlay = gfx_overlay_create(border_rgb, borderw, borderh, 0, 0, 1, GFX_OVERLAY_MID, true);
	    delete [] border_rgb;
	}
    }

    {
	int		 blazew, blazeh;
        u8		*blaze_rgb = gfx_loadbitmap("../gfx/blaze.bmp", blazew, blazeh);

	if (blaze_rgb)
	{
	    blaze_overlay = gfx_overlay_create(blaze_rgb, blazew, blazeh, 960, 0, 1, GFX_OVERLAY_BACK, false);
	    delete [] blaze_rgb;
	}
    }

    ATLAS::createNewWorld();

    // Done world building!
    glbMsg->clear();
    redrawWorld();
    gfx_clearKeyBuffer();
     
    gfx_clearbox(0, 0, SCR_WIDTH, SCR_HEIGHT, '\0', ATTR_NONE);
    bool		done = false;
    bool		didmove;

    BUF			fkeys[12];

    int		worldtests = 0;
    while (0)
    {
	ATLAS::createNewWorld();
	MAP		*world = glbAtlas()->findMap("world");
	int		naltar = world->countTopItems(ITEM_ALTAR);
	worldtests++;
	if (worldtests % 15 == 0)
	    printf("%d", worldtests);
	if (naltar != 4)
	{
	    printf("DAMN");
	}
	assert(naltar == 4);
    }
    
    fadeFromBlack();

    while (!done)
    {
	if (glbVictory)
	{
	    gfx_clearKeyBuffer();
	    fadeToWhite();
	    {
		gfx_overlay_destroy(blaze_overlay);
		gfx_overlay_destroy(border_overlay);
		int		 borderw, borderh;
		u8		*border_rgb = gfx_loadbitmap("../gfx/rough.bmp", borderw, borderh);

		if (border_rgb)
		{
		    gfx_overlay_create(border_rgb, borderw, borderh, 0, 0, 1, GFX_OVERLAY_MID, false);
		    delete [] border_rgb;
		}
	    }
	    readBook(text_lookup("game", "victory"));
	    break;
	}
	if (glbAvatar->hp() == 0)
	{
	    redrawWorld();
	    gfx_getKey(true);

	    fadeToBlack();
	    ATLAS::createNewWorld();

	    // Done world building!
	    redrawWorld();
	    gfx_clearKeyBuffer();

	    readBook(text_lookup("death", "fail"));
	    fadeFromBlack();
	    glbAtlas()->buildFOV(glbAvatar->pos(), 10);
	    redrawWorld();
	}

	glbAtlas()->buildFOV(glbAvatar->pos(), 10);
	redrawWorld();

	didmove = false;

	{
	    int		key;

	    key = gfx_getKey(false);

	    if (key)
	    {
		glbAtlas()->clearMessages();
		glbTalkBox->erase();
		glbTalkBox->clear();

		switch (key)
		{
		    case 'k':
		    case '8':
		    case GFX_KEYUP:
			didmove = glbAvatar->actionBump(DIR::north());
			break;
		    case GFX_KEYDOWN:
		    case '2':
		    case 'j':
			didmove = glbAvatar->actionBump(DIR::south());
			break;
		    case GFX_KEYLEFT:
		    case 'h':
		    case '4':
			didmove = glbAvatar->actionBump(DIR::west());
			break;
		    case GFX_KEYRIGHT:
		    case 'l':
		    case '6':
			didmove = glbAvatar->actionBump(DIR::east());
			break;

#if 0
		    case GFX_KEYF2:
		    {
			MAP *map = glbAtlas()->findMap("cavern1");
			POS p = map->findRandomPassable();
			glbAvatar->moveTo(p);
			break;
		    }
		    case GFX_KEYF3:
		    {
			MAP *map = glbAtlas()->findMap("cavern2");
			POS p = map->findRandomPassable();
			glbAvatar->moveTo(p);
			break;
		    }
		    case GFX_KEYF4:
		    {
			MAP *map = glbAtlas()->findMap("cavern3");
			POS p = map->findRandomPassable();
			glbAvatar->moveTo(p);
			break;
		    }
		    case GFX_KEYF5:
		    {
			MAP *map = glbAtlas()->findMap("cavern4");
			POS p = map->findRandomPassable();
			glbAvatar->moveTo(p);
			break;
		    }
		    case 'R':
		    {
			// rebuild.
			ATLAS::createNewWorld();
			break;
		    }
#endif

		    case '\t':
		    case ' ':
		    {
			glbAvatar->actionShiftPrimary();
			break;
		    }

		    case GFX_KEYF1:
		    case '?':
		    {
			didmove = false;
			readBook(text_lookup("game::keys"));

			break;
		    }

		    case '\n':
		    case '\r':
		    {
			// Invoke selected item.
			didmove = glbAvatar->actionInvokeSelected();
			break;
		    }
		    case 'Q':
		    {
			// Verify they want to quit!
			char buf[100];
			glbTalkBox->getString("Are you sure you wish to quit? ", buf, 100);
			if (buf[0] == 'y' || buf[0] == 'Y')
			    done = true;
			glbTalkBox->clear();
			break;
		    }
		}
	    }
	}
	if (didmove)
	{
	    glbAtlas()->heartbeat();
	}
    }

    gfx_shutdown();
}
